"""The library currently supports the following tasks.

New tasks can be added by inheriting from :class:`SyntaxEvalTask`.
"""

from .lakretz import LakretzTask
from .linzen import LinzenTask
from .marvin import MarvinTask
from .warstadt import WarstadtTask
from .winobias import WinobiasTask
